"Python3 library to handle rectangular data grids (as from R)."

##  GRIDLIB :
##  Python3 data-grid routines :
##  Copyright 2012-2104, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 15/09/2012
##  last revision : 09/01/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
##  import  time, random


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/py3libs")
else:  ##  presume windows
    sys.path.append("c:\\py3libs")
##  in case executed from somewhere else.


##  local libraries :
from globdefs import *
from glib import  numeric, numerize


##  COLS2USE :
def cols2use (paradat,numsonly=True):
    "works out which columns to use in calculations."

    goodcols = set(range(paradat.cols))
    ##  lowtags = [t.lower() for t in paradat.taglist]
    ##  first remove non-numeric columns :
    for vn,varname in enumerate(paradat.taglist):
        nn = paradat.tagdict[varname]
        if numsonly and (nn < paradat.rows):
            goodcols.discard(vn)
    ##  don't remove target column at this stage :
    ##  goodcols.discard(paradat.targcol)
    ##  also remove vars that user asks to omit :
    if paradat.skipvars:
        bits = paradat.skipvars.split(',')  ##  comma-delimited
        for b in bits:
            bb = b.strip()
            if bb in paradat.taglist:
                bcol = paradat.taglist.index(bb)
                goodcols.discard(bcol)
            else:
                print("Warning : skipvar",b,"not found in taglist!")

    return  goodcols  ##  set of usable numeric cols.


##  MATDUMP :
def matdump (fo,dmat):
    "dumps numeric square matrix onto o/p file."

    n = len(dmat)
    lo = 0
    for dvec in dmat:
        lout = vecline(dvec)
        fo.write(lout + NL)
        lo += 1
    assert  lo == n
    
    return  n


##  VECLINE :
def vecline (datavec,decimals=5,dlim=Htab):
    "returns a delimited string from a list of numerics."

    lout = ''
    for v in datavec:
        if type(v) == type(9):
            sv = str(v)
        elif type(v) == type(9.9):
            sv = str(round(v,decimals))
        else:  ##  something wrong ?
            sv = str(v)
        lout = lout + sv + dlim

    return  lout[:-1]  ##  remove trailing delimiter


##  MATREAD :
def matread (filename,paradat,tabchar=Htab):
    "reads data grid with header as first line."

    fi = open(filename,'r',encoding="utf-8")
    ##  crash if failure!
    datlines = fi.readlines()
    h = datlines[0].strip()  ##  presumes header line first
    lc = len(datlines)
    taglist = h.split(tabchar)  ##  assumes tab as delimiter
    cols = len(taglist)
    print(lc,cols)
    ##  print(taglist)
    ##  make dictionary 4 guessing column types :
    tagdict = {}
    for t in taglist:
        tagdict[t] = 0

    datamat = []
    for r in range(1,lc):
        datlist = datlines[r].strip().split(tabchar)
        vals = len(datlist)
        if vals < 1:
            continue  ##  skip blank lines
        if vals < cols:
            print("Warning: dodgy data line :")
            print("items =",vals,"vals = ",datlist)
            return  datamat, taglist  ##  will force early exit
        rowdata = []
        for tp,item in enumerate(datlist):
            v = numerize(item)  ##  most values expected 2b numeric
            if numeric(v):
                rowdata.append(v)
                tagdict[taglist[tp]] += 1
            else:
                rowdata.append(item)  ##  probably string value
                tagdict[taglist[tp]] -= 1
        datamat.append(rowdata)
        
    return  (datamat,taglist,tagdict)


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  VAR2COL :
def var2col (varname,varnames,goodcols=set()):
    "tries to find varname in varnames, only in goodcols if given."

    if not goodcols:
        goodcols = set(range(len(varnames)))
    if varname in varnames:
        vpos = varnames.index(varname)
        if vpos in goodcols:
            return  vpos  ##  position of varname

    return  -1  ##  failure signal


##  BASTATZ :
def bastatz (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = 0.0
    statsack.std1 = statsack.std2 = 0.0
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  VECDIST :
def vecdist (xvec,yvec,p=2):
    "computes Euclidean or p-distance with 1 sweep, Euclid as default."

    n1 = len(xvec) ; n2 = len(yvec)
    n = min(n1,n2)
    ##  what if n1 != n2 ?
    if n < 1:
        return  0.0
    dsum = 0.0
    for j in range(n):
        d = abs(xvec[j]-yvec[j])
        dsum += d**p  ##  p==1 for city-block

    return  dsum ** (1/p)  ##  could divide by n outside ?


##  DISTEST :
def distest (grid):
    "tests whether data grid looks like a distance/dissimilarity matrix."

    rows = len(grid)
    if rows > 0:
        cols = len(grid[0])
    else:
        return  False  ##  could null grid count as sq matrix ??
    if rows != cols:
        return  False
    ##  size okay, test contents :
    for r in range(rows):
        if grid[r][r] != 0.0:
            return  False  ##  must have zero diagonal
        lo = min(grid[r])
        if lo < 0.0:  ##  cd subzero dissims make sense ??
            return  False  ##  negative distances disallowed at present

    return  True
    ##  allows dissimilarity matrices, where i,j locs != j,i locs.


##  COPYGRID :
def copygrid (grid):
    "returns a copy of given data grid."

    newgrid = []
    for r in grid:
        rowdata = []
        for v in r:
            rowdata.append(v)
        newgrid.append(rowdata)

    return  newgrid


##  CENTDIST :
def centdist (paradat,mvec,gvec):
    "computes distance of grid item from midclust."

    xvec = [gvec[gc] for gc in paradat.goodlist]
    if paradat.scaling > 0:
        xvec = rescale(paradat,xvec)
    assert len(xvec) == len(mvec)
    p = 2
    if paradat.distmode == "city":
        p = 1
    dx = vecdist(xvec,mvec,p)

    return  dx


##  MIDDLES :
def middles (paradat,gvec):
    "returns cluster centroids."

    ##  first compute cluster centroids :
    centroid = {} ; cent = {}
    for g in range(paradat.glen):  ##  instance number
        if not (g in paradat.included):
            continue  ##  skip excluded cases
        xv = [paradat.grid[g][j] for j in paradat.goodlist]
        if paradat.scaling > 0:  ##  scaling requested
            xv = rescale(paradat,xv)
        c = gvec[g]  ##  cluster
        if c in centroid:
            cent[c] += 1
            for j,x in enumerate(xv):  ##  cumulative total
                try:
                    centroid[c][j] += x  ##  xv[j]
                except IndexError as xxxx:
                    print(xxxx)
                    print(c,j,len(xv),len(centroid[c]),len(paradat.goodlist))
        else:  ##  first occurrence
            centroid[c] = list(xv)
            cent[c] = 1
    for c in centroid:  ##  mean at present (median later?)
        for j in range(len(paradat.goodlist)):
            centroid[c][j] = centroid[c][j] / cent[c]

    return  centroid


##  RESCALE :
def rescale (xvec,scaling):
    "rescales a numeric vector."

    ##  xvec must be numeric (right thru!)
    xmin = min(xvec) ; xmax = max(xvec)
    xr = (xmax - xmin) or 1.0
    bs = bastatz(xvec)
    vx = []
    for x in xvec:
        if scaling < 1:  ##  unaltered
            vx.append(x)
        elif paradat.scaling == 1:  ##  scaled by range
            xx = (x - xmin) / xr
            vx.append(xx)
        else:  ##  scaled as zedscore
            xx = x - bs.mean
            if bs.std2 > 1e-12:
                xx = xx / bs.std2
            vx.append(xx)

    return  vx

